package ru.amse.bazylevich.faeditor.fautomaton.algorithms.interpreter.impl;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import ru.amse.bazylevich.faeditor.fautomaton.IAutomaton;
import ru.amse.bazylevich.faeditor.fautomaton.IState;
import ru.amse.bazylevich.faeditor.fautomaton.ITransition;
import ru.amse.bazylevich.faeditor.fautomaton.algorithms.interpreter.IInterpreter;
import ru.amse.bazylevich.faeditor.fautomaton.algorithms.interpreter.util.InvalidateAutomatonException;
import ru.amse.bazylevich.faeditor.fautomaton.algorithms.interpreter.util.Way;


public class Interpreter implements IInterpreter {
	
	private final IAutomaton myAutomaton;

	public Interpreter(IAutomaton automaton) {
		myAutomaton = automaton;
	}
	
	
	public Way getWay(String string) throws InvalidateAutomatonException{
		boolean isCorrect = checkString(string);
		if (isCorrect) {
			return getWayForAcceptedString(string, isCorrect);
		}
		return getWayForNotAcceptedString(string, isCorrect);
	}
	
	
	private boolean checkString(CharSequence string) throws InvalidateAutomatonException {
		if (myAutomaton == null) {
			throw new InvalidateAutomatonException("Null");
		}
		IState initialState = myAutomaton.getInitialState();
		//!!!!!!!!!!!
		if ((initialState == null)) {// || (string.length() == 0)) {
			throw new InvalidateAutomatonException("This automaton hasn't initial state");
		} 
		return checkEndOfString(string, 0, initialState);
	}

	
	private boolean checkEndOfString(CharSequence string, int n, IState state) {
		if (n  == string.length() ) {
			return myAutomaton.getFinalStates().contains(state);
		} 
		Set<IState> nextStates = getNextStates(string.charAt(n), state);
		for (IState nextState : nextStates) {
			if (checkEndOfString(string, n + 1, nextState)) {
				return true;
			}
		}
		return false;
	}
	
	private Set<IState> getNextStates(char c, IState state) {
		Set<IState> setStates = new HashSet<IState>();
		for (ITransition transition : state.getTransitions()) {
			if (transition.getCondition().getSymbols().contains(c)) {
				setStates.add(transition.getEnd());
			}
		}
		return setStates;
	}
	
	private Set<ITransition> getTransitions(char c, IState state) {
		Set<ITransition> setTransitions = new HashSet<ITransition>();
		for (ITransition transition : state.getTransitions()) {
			if (transition.getCondition().getSymbols().contains(c)) {
				setTransitions.add(transition);
			}
		}
		return setTransitions;
	}
	
	
	
	private Way getWayForNotAcceptedString(CharSequence string, boolean isCorrect) {
		List<ITransition> bestWay = new LinkedList<ITransition>();
		if (string.length() != 0) {
			List<ITransition> currentWay = new LinkedList<ITransition>();
			Set<ITransition> setTransition = getTransitions(string.charAt(0), myAutomaton.getInitialState());
			if (setTransition.size() != 0) {
				for (ITransition nextTransition : setTransition) {
					makeWayForNotAcceptedString(string, 1, nextTransition, currentWay, bestWay);
					((LinkedList<ITransition> )currentWay).removeLast();
				}
			} 
		}
		return new Way(bestWay, makeStatesWay(bestWay), isCorrect);
	}
	
	private void makeWayForNotAcceptedString(CharSequence string, int n, ITransition transition, List<ITransition> currentWay, List<ITransition> bestWay) {
		currentWay.add(transition);
		if (currentWay.size() > bestWay.size()) {
			bestWay.clear();
			bestWay.addAll(currentWay);
		}
		if (n == string.length()) {
			return;
		}
		Set<ITransition> setTransition = getTransitions(string.charAt(n), transition.getEnd());
		if (setTransition.size() != 0) {
			for (ITransition nextTransition : setTransition) {
				makeWayForNotAcceptedString(string, n + 1, nextTransition, currentWay, bestWay);
				((LinkedList<ITransition> )currentWay).removeLast();
			}
		} 
	}

	private Way getWayForAcceptedString(CharSequence string, boolean isCorrect) {
		List<ITransition> currentWay = new LinkedList<ITransition>();
		if (string.length() != 0) {
			Set<ITransition> setTransition = getTransitions(string.charAt(0), myAutomaton.getInitialState());
			for (ITransition transition : setTransition) {
				if (makeWayForAcceptedString(string, 1, transition, currentWay)) {
					return new Way(currentWay,  makeStatesWay(currentWay), isCorrect);
				}
			}
		}
		return new Way(currentWay,  makeStatesWay(currentWay), isCorrect);
	}
	
	private boolean makeWayForAcceptedString(CharSequence string, int n, ITransition transition, List<ITransition> currentWay) {
		if (n == string.length()) {
			if (myAutomaton.getFinalStates().contains(transition.getEnd())) {
				currentWay.add(0, transition);
				return true;
			}
			return false;
		}
		Set<ITransition> setTransition = getTransitions(string.charAt(n), transition.getEnd());
		for (ITransition nextTransition : setTransition) {
			if (makeWayForAcceptedString(string, n + 1, nextTransition, currentWay)) {
				currentWay.add(0, transition);
				return true;
			}
		}
		return false;
	}
	
	
	private List<IState> makeStatesWay(List<ITransition> way) {
		List<IState> statesWay = new LinkedList<IState>();
		statesWay.add(myAutomaton.getInitialState());
		for (ITransition transition : way) {
			statesWay.add(transition.getEnd());
		}
		if (statesWay.size() == 0) {
			statesWay.add(myAutomaton.getInitialState());
		}
		return statesWay;
	}
}
